package org.japura.task;

import java.util.AbstractQueue;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

/**
 * <P>
 * Copyright (C) 2011-2013 Carlos Eduardo Leite de Andrade
 * <P>
 * This library is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 * <P>
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * <P>
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <A
 * HREF="www.gnu.org/licenses/">www.gnu.org/licenses/</A>
 * <P>
 * For more information, contact: <A HREF="www.japura.org">www.japura.org</A>
 * <P>
 * 
 * @author Carlos Eduardo Leite de Andrade
 */
public class FifoBlockingQueue extends AbstractQueue<Runnable> implements
	BlockingQueue<Runnable>{

  private ArrayList<Runnable> queue = new ArrayList<Runnable>();

  private final ReentrantLock takeLock = new ReentrantLock();

  private final Condition notEmpty = takeLock.newCondition();

  private final ReentrantLock putLock = new ReentrantLock();

  private void signalNotEmpty() {
	takeLock.lock();
	try {
	  notEmpty.signal();
	} finally {
	  takeLock.unlock();
	}
  }

  private void enqueue(Runnable e) {
	FutureTaskWrapper<?> fw = (FutureTaskWrapper<?>) e;
	if (fw.getTask().isPriority()) {
	  queue.add(0, e);
	  fw.getTask().clearPriority();
	} else {
	  queue.add(e);
	}
  }

  private Runnable dequeue() {
	if (queue.size() > 0) {
	  return queue.remove(0);
	}
	return null;
  }

  void fullyLock() {
	putLock.lock();
	takeLock.lock();
  }

  void fullyUnlock() {
	takeLock.unlock();
	putLock.unlock();
  }

  @Override
  public int size() {
	return queue.size();
  }

  @Override
  public int remainingCapacity() {
	return Integer.MAX_VALUE - size();
  }

  @Override
  public void put(Runnable e) throws InterruptedException {
	if (e == null)
	  throw new NullPointerException();
	putLock.lockInterruptibly();
	try {
	  enqueue(e);
	} finally {
	  putLock.unlock();
	}
	if (size() > 0) {
	  signalNotEmpty();
	}
  }

  @Override
  public boolean offer(Runnable e, long timeout, TimeUnit unit)
	  throws InterruptedException {
	return offer(e);
  }

  @Override
  public boolean offer(Runnable e) {
	if (e == null) {
	  throw new NullPointerException();
	}
	if (size() == Integer.MAX_VALUE) {
	  return false;
	}
	putLock.lock();
	try {
	  if (size() < Integer.MAX_VALUE) {
		enqueue(e);
	  }
	} finally {
	  putLock.unlock();
	}
	if (size() > 0) {
	  signalNotEmpty();
	}
	return size() > 0;
  }

  @Override
  public Runnable take() throws InterruptedException {
	Runnable r;
	takeLock.lockInterruptibly();
	try {
	  while (size() == 0) {
		notEmpty.await();
	  }
	  r = dequeue();
	  if (size() > 0) {
		notEmpty.signal();
	  }
	} finally {
	  takeLock.unlock();
	}
	return r;
  }

  @Override
  public Runnable poll(long timeout, TimeUnit unit) throws InterruptedException {
	return poll();
  }

  @Override
  public Runnable poll() {
	if (size() == 0) {
	  return null;
	}
	Runnable r = null;
	takeLock.lock();
	try {
	  if (size() > 0) {
		r = dequeue();
		if (size() > 0) {
		  notEmpty.signal();
		}
	  }
	} finally {
	  takeLock.unlock();
	}
	return r;
  }

  @Override
  public Runnable peek() {
	if (size() == 0) {
	  return null;
	}
	takeLock.lock();
	try {
	  if (size() > 0) {
		return queue.get(0);
	  }
	  return null;
	} finally {
	  takeLock.unlock();
	}
  }

  @Override
  public boolean remove(Object o) {
	if (o == null) {
	  return false;
	}
	fullyLock();
	try {
	  return queue.remove(o);
	} finally {
	  fullyUnlock();
	}
  }

  @Override
  public Object[] toArray() {
	fullyLock();
	try {
	  return queue.toArray(new Object[0]);
	} finally {
	  fullyUnlock();
	}
  }

  @SuppressWarnings("unchecked")
  @Override
  public <T> T[] toArray(T[] a) {
	fullyLock();
	try {
	  for (int i = 0; i < a.length && i < size(); i++) {
		a[i] = (T) queue.get(i);
	  }
	  return a;
	} finally {
	  fullyUnlock();
	}
  }

  @Override
  public void clear() {
	fullyLock();
	try {
	  queue.clear();
	} finally {
	  fullyUnlock();
	}
  }

  @Override
  public int drainTo(Collection<? super Runnable> c) {
	return drainTo(c, Integer.MAX_VALUE);
  }

  @Override
  public int drainTo(Collection<? super Runnable> c, int maxElements) {
	if (c == null)
	  throw new NullPointerException();
	if (c == this)
	  throw new IllegalArgumentException();
	takeLock.lock();
	try {
	  int n = Math.min(maxElements, size());
	  for (int i = 0; i < n; i++) {
		Runnable r = queue.get(i);
		c.add(r);
	  }
	  return n;
	} finally {
	  takeLock.unlock();
	}
  }

  @Override
  public Iterator<Runnable> iterator() {
	return queue.iterator();
  }

}